def _load_df(meta):
    import pandas as pd
    import json

    with open(meta) as f:
        meta = json.load(f)

    g  = []
    t  = []
    s  = []
    for identifier in meta['identifiers']:
        t.append(meta['search'][identifier])

        for group, identifiers in meta['group'].items():
            if identifier in identifiers:
                g.append(group)
                continue

        s.append('train' if any(identifier in meta['group'][group] for group in meta['Train']) else 'val')

    df = pd.DataFrame()
    df['identifier'] = meta['identifiers']
    df['group']      = g
    df['target']     = t
    df['train/val']  = s
    df.index         += 1

    return df
    
def _count(meta):
    import pandas as pd
    df    = _load_df(meta)
    count = pd.DataFrame(dtype = int)
    for target in sorted(set(df['target'])):
        for group in sorted(set(df['group'])):
            count.loc[target, group] = ((df['target'] == target) & (df['group'] == group)).sum()

    if 'train' in count.columns:
        count.columns  = [column.title() for column in count.columns]
    else:
        count['Train'] = count[list(meta['Train'])].sum(axis = 1)
        count['Val'  ] = count[list(meta['Val'  ])].sum(axis = 1)

    count['Total'] = count['Train'] + count['Val']
    count.loc['Total'] = count.sum(axis = 0)

    return count

def _print(func):
    def inner(meta):
        print(f'genolearn command : print metadata {func.__name__}')
        print(f'metadata          : {meta.split("/")[-1]}', '\n')
        print(func(meta))
    return inner
    
@_print    
def count(meta):
    df = _count(meta).applymap(int)
    return df

@_print   
def proportion(meta):
    df = _count(meta)
    df.iloc[:-1] /= df.iloc[:-1].sum(axis = 0)
    df.iloc[-1]   = df.iloc[-1] / df.iloc[-1,-1]
    df *= 100
    return df.round(0).applymap(int)

@_print   
def head(meta, num = 10):
    df = _load_df(meta)
    if (df['group'] == df['train/val']).all():
        df.drop('group', axis = 1, inplace = True)
    return df.head(num)
    
@_print   
def tail(meta, num = 10):
    df = _load_df(meta)
    if (df['group'] == df['train/val']).all():
        df.drop('group', axis = 1, inplace = True)
    return df.tail(num)

@_print   
def sample(meta, num = 10):
    df = _load_df(meta)
    if (df['group'] == df['train/val']).all():
        df.drop('group', axis = 1, inplace = True)
    return df.sample(num)



def base_feature_selection(method, dataloader, init, loop, post):
    """
    base feature selection function

    Parameters
    ----------
        method : str
            Method name.

        dataloader : str
            genolearn.dataloader.DataLoader object.

        init : str
            Initialise function to generate and the associated ``args`` and ``kwargs`` for ``inner_loop`` and ``outer_loop`` functions.

        inner_loop : str
            Inner loop function to be executed on a given (x, y) pair.
        
        outer_loop : str
            Outer loop function to be executed for each value in ``values``.
    """
    from   genolearn.logger import msg, Waiting
    args, kwargs = init(dataloader)
    n            = sum(map(len, (dataloader.meta['group'][group] for group in dataloader.meta['Train'])))
    for i, (x, label) in enumerate(dataloader.generator('Train'), 1):
        msg(f'{method} : {i:,d} of {n:,d}', inline = True)
        loop(i, x, label, 'Train', *args, **kwargs)
    
    with Waiting(f'{method} : {i:,d} (computing scores)', f'{method} : {i:,d} (computed scores)'):
        ret = post(i, 'Train', *args, **kwargs)

    return ret

def feature_selection(name, meta, method, module, log):

    from   genolearn.logger     import msg, Writing
    from   genolearn.dataloader import DataLoader
    from   genolearn            import utils, working_directory

    import numpy  as np
    import os

    dataloader = DataLoader(meta, working_directory)
    
    os.makedirs('feature-selection', exist_ok = True)

    variables  = {}
    with open(os.path.expanduser(module)) as f:
        exec(f.read(), {}, variables)

    save_path    = os.path.join('feature-selection', name)
    funcs        = ['init', 'loop', 'post']

    for name in funcs:
        assert name in variables
        
    params       = {func : variables.get(func) for func in funcs}
    scores       = base_feature_selection(method, dataloader, **params)

    with Writing(save_path, inline = True):
        np.savez_compressed(save_path, scores)
        os.rename(f'{save_path}.npz', save_path)

    utils.create_log(log, 'feature-selection')

    msg(f'executed "genolearn feature-selection"')



def evaluate(data_config, feature_selection, values, encoder, num_features, binary, output):

    from   genolearn.logger import msg

    from   genolearn.dataloader import DataLoader

    import joblib
    import pandas as pd

    import os

    dtype = bool if binary else None

    for model, pre in zip(['model.pickle', 'full-model.pickle'], ['', 'full-']):
        model       = joblib.load(os.path.join('..', model))
        dataloader  = DataLoader(**data_config)

        features = dataloader.load_feature_selection(feature_selection).argsort()[:num_features]

        X   = dataloader.load_X(*values, features = features, dtype = dtype)

        dataloader._encoder = encoder

        df  = pd.DataFrame(index = dataloader._check_identifiers(values))
        df.index.name = 'identifier'

        df['hat'] = dataloader.decode(model.predict(X))
        
        if hasattr(model, 'predict_proba'):
            prob = model.predict_proba(X)
            for i, name in enumerate(encoder):
                df[f'P({name})'] = prob[:,i]
        
        df.to_csv(f'{pre}{output}')

    msg('executed "genolearn evaluate"')



from   genolearn import working_directory
import os
import json
import scipy.sparse
import numpy  as np
import gzip
import warnings

warnings.filterwarnings('ignore')

class DataLoader():

    def __init__(self, meta_file, working_dir = working_directory):
        
        if working_dir is None:
            raise Exception('execute GenoLearn\'s setup command first')

        if not os.path.exists(working_dir) or '.genolearn' not in os.listdir(working_dir):
            raise Exception(f'"{working_dir}" not a valid working directory')

        path  = os.path.join(working_dir, 'meta')
        if not os.path.exists(path):
            raise Exception('no preprocessed metadata to load')
        
        valid = os.listdir(path)

        if len(valid) == 0:
            raise Exception(f'no preprocessed metadata in {path}')
            
        if meta_file not in valid:
            raise Exception(f'"{meta_file}" not found in "{path}" - expect one of' + '\n  •'.join([''] + valid))

        self.working_dir    = working_dir
        self.preprocess_dir = os.path.join(working_dir, 'preprocess')
        self.data_dir       = os.path.join(self.preprocess_dir, 'array')
        
        with open(os.path.join(working_dir, 'meta', meta_file)) as f:
            self.meta = json.load(f)
            self.c    = len(set(self.meta['targets']))

        with open(os.path.join(self.preprocess_dir, 'info.json')) as f:
            info = json.load(f)
            self.n, self.m = info['n'], info['m']

    def _load_X(self, identifier, features, dtype):

        npz  = os.path.join(self.data_dir, f'{identifier}.npz')

        arr, = np.load(npz).values()

        if features is not None:
            arr = arr[features]
        
        return arr.astype(dtype) if dtype else arr

    def _load_Y(self, identifier):
        return self.meta['search'][identifier]

    def _check_identifiers(self, identifiers):
        ret = []
        
        for identifier in identifiers:

            # unlabelled data
            if identifier == 'unlabelled':
                for identifier in os.listdir(self.data_dir):
                    identifier = identifier.replace('.npz', '')
                    if identifier not in self.meta['identifiers']:
                        ret.append(identifier)
                    
            # check if it is a group and append entire associated identifiers
            elif identifier in self.meta['group']:
                ret += self.meta['group'][identifier]
            
            # check if it is Train / Val and append entire associated identifiers
            elif identifier in ['Train', 'Val']:
                for group in self.meta[identifier]:
                    ret += self.meta['group'][group]

            # append the identifier as is
            else:
                ret.append(identifier)

        return ret

    def load_X(self, *identifiers, features = None, dtype = None):
        identifiers = self._check_identifiers(identifiers)
        features    = np.array(features)
        m           = self.m if features is None else features.sum() if features.dtype == np.bool_ else len(features)
        X           = np.zeros((len(identifiers), m), dtype = dtype)
        for i, identifier in enumerate(identifiers):
            X[i] = self._load_X(identifier, features, dtype)
        return X

    def load_Y(self, *identifiers):
        """
        Returns
        -------
        Y : numpy.ndarray
        """
        identifiers = self._check_identifiers(identifiers)
        return np.array(list(map(self._load_Y, identifiers)))

    def load_train_val_identifiers(self, min_count = 0, target_subset = None):
        train   = self._check_identifiers(self.meta['Train'])
        val     = self._check_identifiers(self.meta['Val'])

        y_train = self.load_Y(*train)
        y_val   = self.load_Y(*val)

        unique, arg = np.unique(y_train, return_inverse = True)
        dummies     = np.eye(len(unique))[arg]

        label_counts = dummies.sum(axis = 0)
        labels       = unique[label_counts >= min_count]

        if target_subset:
            labels = [label for label in labels if label in target_subset]
        
        self._encoder = {label : i for i, label in enumerate(labels)}

        identifiers_train = np.array(train)[np.isin(y_train, labels)]
        identifiers_val   = np.array(val  )[np.isin(y_val  , labels)]

        return identifiers_train, identifiers_val

    def load_train_val(self, features = None, dtype = None, min_count = 0, target_subset = None):
        self.identifiers_train, self.identifiers_val = self.load_train_val_identifiers(min_count, target_subset)

        Y_train = self.encode(self.load_Y(*self.identifiers_train))
        Y_val   = self.encode(self.load_Y(*self.identifiers_val  ))

        X_train = self.load_X(*self.identifiers_train, features = features, dtype = dtype)
        X_val   = self.load_X(*self.identifiers_val  , features = features, dtype = dtype)

        return X_train, Y_train, X_val, Y_val
    
    def load_identifiers(self, min_count = 0, target_subset = None):
        train   = self._check_identifiers(self.meta['Train'] + self.meta['Val'])

        y_train = self.load_Y(*train)

        unique, arg = np.unique(y_train, return_inverse = True)
        dummies     = np.eye(len(unique))[arg]

        label_counts = dummies.sum(axis = 0)
        labels       = unique[label_counts >= min_count]

        if target_subset:
            labels = [label for label in labels if label in target_subset]
        
        self._encoder = {label : i for i, label in enumerate(labels)}

        identifiers = np.array(train)[np.isin(y_train, labels)]

        return identifiers

    def load(self, features = None, dtype = None, min_count = 0, target_subset = None):

        self.identifiers = self.load_identifiers(min_count, target_subset)

        Y = self.encode(self.load_Y(*self.identifiers))
        X = self.load_X(*self.identifiers, features = features, dtype = dtype)

        return X, Y

    def encode(self, Y):
        return np.vectorize(lambda value : self._encoder[value])(Y)

    def decode(self, Y):
        decoder = {value : key for key, value in self._encoder.items()}
        return np.vectorize(lambda value : decoder[value])(Y)

    def features(self, indices = None):
        if '_features' not in self.__dict__:
            with gzip.open(os.path.join(self.preprocess_dir, 'features.txt.gz')) as g:
                self._features = g.read().decode().split()
        return self._features if indices is None else [self._features[i] for i in indices]

    def load_feature_selection(self, file):
        ret, = np.load(os.path.join(self.working_dir, 'feature-selection', file), allow_pickle = True).values()
        return ret

    def generator(self, *identifiers, features = None, dtype = None):
        """
        Iteratively yields an x, y pair from the method ``load``.
        """
        identifiers = self._check_identifiers(identifiers)

        for identifier in identifiers:
            yield self._load_X(identifier, features, dtype), self.load_Y(identifier)[0]



__version__ = '0.1.2'

import os

ls = os.listdir()
wd = True if '.genolearn' in ls else False

path = os.path.join(os.path.dirname(__file__), 'wd')
if wd:
    working_directory = os.path.abspath('.')
    with open(path, 'w') as f:
        f.write(working_directory)
elif os.path.exists(path):
    with open(path) as f:
        working_directory = f.read()
        if not (os.path.exists(working_directory) and '.genolearn' in os.listdir(working_directory)):
            working_directory = None
else:
    working_directory = None


ls = os.listdir(working_directory if working_directory and os.path.exists(working_directory) else '.')

class Path():

    def add(self, func):
        name = func.__name__
        if name.startswith('_'):
            name = name[1:]
        self.__dict__[name] = func

path = Path()

def add(func):
    path.add(func)
    return lambda *args, **kwargs : func(*args, **kwargs)
    
@add
def join(*args):
    return os.path.join(working_directory, *args)

@add
def join_from(path):
    return lambda *args : os.path.join(path, *args)

@add
def listdir(path = '.', *args):
    if working_directory and os.path.exists(working_directory):
        path = join(path, *args)
        return os.listdir(path) if os.path.exists(path) else []
    return []

@add
def expanduser(path, inverse = False):
    return path.replace('~', os.path.expanduser('~')) if inverse else path.replace(os.path.expanduser('~'), '~')   

@add
def _open(path, mode = 'r'):
    return open(join(path), mode)

@add
def exists(path):
    return os.path.exists(os.path.expanduser(path))

def get_active():
    try:
        if working_directory:
            path = join('.genolearn')
            if os.path.exists(path):
                import json
                with open(path) as f:
                    return json.load(f)
    except:
        ...



def init(dataloader):
    """
    Initialises statistics for the Fisher Score computation.
    """
    import numpy as np

    # encoding from class label to integer
    encode = {c : i for i, c in enumerate(set(dataloader.meta['targets']))}

    # class label count
    n  = np.zeros(dataloader.c)

    # global sum
    sg = np.zeros(dataloader.m)

    # class label sum
    s1 = np.zeros((dataloader.m, dataloader.c))

    # class label sum of squares
    s2 = np.zeros((dataloader.m, dataloader.c))

    args   = (encode, n, sg, s1, s2)  # encoder, counts, sum, by class sum, by class sum of squares
    kwargs = {}                       # no kwargs

    return args, kwargs

def loop(i, x, label, value, *args, **kwargs):
    """
    Incrementally updates count, global, and by class label statistics
    """

    encode, n, sg, s1, s2 = args
        
    y        = encode[label]

    # convert to binary 0 -> 0, {1, 2, 3, ...} -> 1
    x        = x > 0

    # increase count of label
    n[y]    += 1

    # increase global sum
    sg      += x

    # increase class label sum
    s1[:,y] += x

    # increase class label sum of squares
    s2[:,y] += x ** 2

def post(i, value, *args, **kwargs):
    """
    Computes the Fisher Score using statistics stored in ``*args``.
    """
    import numpy as np
    
    encode, n, sg, s1, s2 = args

    # convert global sum to global mean
    mu  = sg / n.sum()

    # convert to first and second moments
    m1  = np.divide(s1, n, where = n > 0)
    m2  = np.divide(s2, n, where = n > 0)

    # compute D and S as per www.genolearn.readthedocs.io/usage/feature-selection.html
    D   = np.square(m1 - mu.reshape(-1, 1)) # broadcast second dimension ((m, c) - (m, 1))
    S   = (m2 - np.square(m1))

    # numerator and denominator expressions for Fisher Score
    num = D @ n
    den = S @ n

    S   = np.divide(num, den, where = den > 0)

    return -S # return negative scores such that argsort returns largest to smallest



from   genolearn.logger     import Writing

from   sklearn.ensemble     import AdaBoostClassifier, RandomForestClassifier
from   sklearn.linear_model import LogisticRegression

import numpy as np
import os

class FeatureImportance():
    """
    Feature Importance Class

    Parameters
    ----------
        model : object
            Fitted scikit-learn model.

    Attributes
    ----------

        feature_score : numpy.ndarray
            Numpy array (dtype = float) of feature importance scores.

    Methods
    -------

        ranked_features: numpy.ndarray
            Numpy array (dtype = int) of ranked features.
    """
    def __init__(self, model):
        self.model = model
        self._detailed_scores, self._aggregate_scores = _assign(model)
        
    @property
    def detailed_scores(self):
        return self._detailed_scores(self.model)

    @property
    def aggregate_scores(self):
        return self._aggregate_scores(self.model)  

def _assign(model):
    if isinstance(model, LogisticRegression):
        return _logistic_detailed_scores, _logistic_aggregate_scores
    elif isinstance(model, (RandomForestClassifier, AdaBoostClassifier)):
        return _ensemble_detailed_scores, _ensemble_aggregate_scores
    else:
        raise Exception()

def _logistic_detailed_scores(model):
    return model.coef_

def _logistic_aggregate_scores(model, fun = np.absolute, **kwargs):
    return fun(_logistic_detailed_scores(model)).mean(axis = 0)

def _ensemble_detailed_scores(model):
    return model.feature_importances_

def _ensemble_aggregate_scores(model, **kwargs):
    return _ensemble_detailed_scores(model)

def feature_importance(meta, feature_selection, num_features, output):

    from   genolearn.dataloader import DataLoader
    from   genolearn.logger     import Writing, msg

    import pickle

    import pandas as pd
    import numpy  as np

    dataloader = DataLoader(meta)

    selection  = dataloader.load_feature_selection(feature_selection).argsort()

    features   = dataloader.features(selection[:num_features])

    for model, pre in zip(['model.pickle', 'full-model.pickle'], ['', 'full-']):
        with open(os.path.join(os.path.dirname(output), model), 'rb') as f:
            model = pickle.load(f)

        importance = FeatureImportance(model)

        d_scores   = importance.detailed_scores
        a_scores   = importance.aggregate_scores

        ranks      = a_scores.argsort()[::-1]

        features   = np.array(features)

        npz        = os.path.join(output, f'{pre}importance.npz')
        csv        = os.path.join(output, f'{pre}importance-rank.csv')

        os.makedirs(output, exist_ok = True)

        with Writing(npz, inline = True):
            np.savez(npz, features = features, ranks = ranks, detailed_scores = d_scores, aggregate_scores = a_scores)

        with Writing(csv, inline = True):
            df = pd.DataFrame()
            df['features'] = features[ranks]
            df['aggregate_score'] = a_scores[ranks]
            df.to_csv(csv, index  = False)

    msg(f'executed "genolearn feature-importance"')